module("L_VeraConnectWWN1", package.seeall)
-- andrei : {"access_token":"c.CzL215BijVuzFgznRCW9QKvA0eblyxSMI9ad7klSX87vkpDyI0XxiftueUPWqDim9ejCCrYJ9r0WCHhJz16clWAxoLTqhcbXeYT45kzbeJ8gC5UwLIxdrc0v2teyuVhnSz6ZLmTJ2TEqEXIm","expires_in":315360000}
-- florin : {"access_token":"c.7zKZ8yzOgDWUKGUTeBKz1JOPRAM6uWOse7Eb6jnfcKTT23hyOLKV5pv3vafKRp6wHlRHkVjknRABDa4NKnxAJNl6lq4GIoJPXIX4mby3ubcMprxqrM2GFN0EGb2WdWQwtsTblXLVgkKvGG2Y","expires_in":315360000}

--require("socket")
local ltn12 = require("ltn12")
local dkjson = require ("dkjson")
local https = require("ssl.https")
local http = require ("socket.http")
--local mime = require("mime")
-- Flags
local DEBUG_MODE = false
local FLAGS = {
	
}

-- Constants
local DEVICE_FILES = {
	TEMPERATURE_SENSOR 	= "D_TemperatureSensor1.xml",
	THERMOSTAT			= "D_HVAC_ZoneThermostat1.xml",
	HUMIDITY			= "D_HumiditySensor1.xml",
	SMOKE_SENSOR		= "D_SmokeSensor1.xml",
	BINARY_LIGHT		= "D_BinaryLight1.xml"
}

local DEVICE_TYPES = {
	TEMPERATURE_SENSOR 	= "urn:schemas-micasaverde-com:device:TemperatureSensor:1",
	THERMOSTAT			= "urn:schemas-upnp-org:device:HVAC_ZoneThermostat:1",
	HUMIDITY			= "urn:schemas-micasaverde-com:device:HumiditySensor:1",
	SMOKE_SENSOR		= "urn:schemas-micasaverde-com:device:SmokeSensor:1",
	BINARY_LIGHT		= "urn:schemas-upnp-org:device:BinaryLight:1"
}

local SID = {
	NEST      		= "urn:micasaverde-com:serviceId:VeraConnectWWN1",
	SES        		= "urn:micasaverde-com:serviceId:SecuritySensor1",
	TEMP			= "urn:upnp-org:serviceId:TemperatureSensor1",
	HVAC_UOM		= "urn:upnp-org:serviceId:HVAC_UserOperatingMode1",
	HVAC_OS			= "urn:micasaverde-com:serviceId:HVAC_OperatingState1",
	HVAC_FOM		= "urn:upnp-org:serviceId:HVAC_FanOperatingMode1",
	HVAC_HEAT 		= "urn:upnp-org:serviceId:TemperatureSetpoint1_Heat",
	HVAC_COOL 		= "urn:upnp-org:serviceId:TemperatureSetpoint1_Cool",
	HADEVICE 		= "urn:micasaverde-com:serviceId:HaDevice1",
	ZWN 	    	= "urn:micasaverde-com:serviceId:ZWaveNetwork1",
	SWP      		= "urn:upnp-org:serviceId:SwitchPower1"
}

local TASK = {
	ERROR       = 2,
	ERROR_ALARM = -2,
	ERROR_STOP  = -4,
	SUCCESS     = 4,
	BUSY        = 1
}
-- LastUpdate
local DISPLAY_SECONDS = 20
local POLLING_RATE = 5 -- 30 --"PollFrequency"

-- Globals

local lug_device = nil
local g_appendPtr -- The pointer passed to luup.chdev.append
local g_taskHandle = -1
local g_lastTask = os.time() -- The time when the status message was last updated.
local g_username = ""
local g_password = ""
local g_fileSize = 0
local g_maxFileSize = 5 * 1024 * 1024
local g_thermostats = {
	-- .panelID = ""
	-- .name = ""
	-- .devId
}
local g_smoke = {

}
local g_structures = {

}

local ipAddress
local ipPort
local NEST_UA = "Nest/1.1.0.10 CFNetwork/548.0.4"
local nest_token = ""
local nest_ClientID = "5b1585a5-b64e-4759-903e-895ac9373245"
local nest_Secret = "noR9PTaBOHx6kONdieaClyyBX"
local g_URL = "https://developer-api.nest.com"
local temperatureScale = ""
-- 7W69Q34L
---------------------------------------------------------
-----------------Generic Utils---------------------------
---------------------------------------------------------
local function log(functionName, text)
	luup.log ("(Vera Connect WWN)::(" .. functionName .. ") : " .. text)
end

local function debug (functionName, text)
	if (DEBUG_MODE == true) then
		log(functionName, text)
	end
end

local function displayMessage (text, mode)
	if (mode == TASK.ERROR_ALARM or mode == TASK.ERROR_STOP) then
		luup.task(text, TASK.ERROR, "Vera Connect WWN Plugin", g_taskHandle)
		if (mode == TASK.ERROR_STOP) then
			luup.set_failure(true, lug_device)
		end
		return
	end
	luup.task(text, mode, "Vera Connect WWN Plugin", g_taskHandle)
	-- Set message timeout.
	g_lastTask = os.time()
	luup.call_delay("clearTask", DISPLAY_SECONDS)
end

local function UrlEncode (s)
	s = s:gsub("\n", "\r\n")
	s = s:gsub("([^%w])", function (c)
							  return string.format("%%%02X", string.byte(c))
						  end)
	return s
end

local function getTemperatureScale()
	local code, data = luup.inet.wget("http://localhost:3480/data_request?id=lu_sdata")
	if (code == 0) then
		data = dkjson.decode(data)
	end
	temperatureScale = ((code == 0) and (data ~= nil) and (data.temperature ~= nil)) and string.lower(data.temperature) or "f"
end

local function getNestDeviceID(device, data_structure)
	for k,v in pairs (data_structure) do
		if v.vera_id == device then
			return v.device_id
		end
	end
	return
end

local function nestRequest(method, headers, URL, data)
	--debug("nestRequest","method = " .. method)
	--debug("nestRequest","headers = " .. tostring(headers))
	--debug("nestRequest","URL = " .. URL)
	--debug("nestRequest","URL = " .. URL)
	local curlCommand = ""
	if method == "GET" or method == "POST" then
		curlCommand = "curl -k -L -X " .. method .. " " .. headers .. " " .. URL
	elseif method == "STREAM" then
		curlCommand = "curl -N -k -v -L " .. URL .. " " .. headers .. " > /tmp/NestOutput.txt &" 
	else
		curlCommand = "curl -k -v -L -X " .. method .. " " .. URL .. " " .. headers .. " " .. data
	end
	--debug("nestRequest", "curlCommand = " .. tostring(curlCommand))
	local stdout = io.popen(curlCommand)
	local response = stdout:read("*a")
	stdout:close()	
		--debug("nestRequest", "curlCommand Response = [" .. tostring(response) .. "]")
	
	if method == "STREAM" then
		local fileCreated = io.open("/tmp/NestOutput.txt")
		if fileCreated then
			return true
		else
			return
		end
	end
	
	if response and response ~= "" then
		local json_string = response
		local json_response = dkjson.decode(response)
		local data = {}
		if json_response.error then
			debug("nestRequest", "error = " .. tostring(json_response.error))
			debug("nestRequest", "description = " .. tostring(json_response.error_description))
			return
		else
			return json_response, json_string
		end
		--[[
		for k,v in pairs(response) do
			log("get_token","key = [" .. tostring(k) .. "] and value = [" .. v .. "]")
		end
		-- ]]
		return response
	else
		debug("nestRequest", "Empty response!")
		return
	end
end

local function get_token()
end

local function clearNestOutputFile()
	local stdout = io.popen("ps | grep 'curl -N -k -v -L https://developer-api.nest.com/devices.json?auth=' | grep -v 'grep'")
	local response = stdout:read("*a")
	stdout:close()
	response = response:gmatch("(%d+) root")
	for v in response do
		debug("clearNestOutputFile", "Killed process " .. v)
		os.execute("kill " .. v)
	end
	os.execute("rm /tmp/NestOutput.txt")
end

local function startStreaming()
	clearNestOutputFile()
	local url = g_URL .. "/devices.json?auth=" .. nest_token
	local header = "-H \"Accept: text/event-stream\" "
	local response, response_string = nestRequest("STREAM", header, url, "")
	if response then
		debug("startStreaming", "Streaming started successful!")
	else
		--debug("startStreaming", "Nest Data = " .. response_string)
		debug("startStreaming", "ERROR : Could not start Stream!")
	end
end

function getStreamEvents()
	debug("getStreamEvents", "Getting Stream Events ...")
	local lfs = require "lfs"
	local fileSize = lfs.attributes("/tmp/NestOutput.txt", "size")
	debug("getStreamEvents", "File size is = [" .. tostring(fileSize) .. "]")
		
	-- check for new stram events
	if fileSize then
		if fileSize > g_fileSize then
			local file = io.open("/tmp/NestOutput.txt")
			file:seek("set", g_fileSize)
			local newEvent = file:read("*a")
			file:close()
			newEvent = newEvent:match("data: (.*)}")
			debug("getStreamEvents", "newEvent = [" .. tostring(newEvent) .. "]")
			g_fileSize = fileSize
			if newEvent and newEvent ~= "null" then
				local response = dkjson.decode(newEvent .. "}")
				local therm = response.data.thermostats
				--[[
				for key,val in pairs(therm) do
					for k,v in pairs(val) do
						debug("getStreamEvents","therm[" .. key .."][" .. tostring(k) .. "]=[" .. tostring(v) .. "]")
					end
				end	
				]]
				if therm then
					for i=1,#g_thermostats do
						if g_thermostats[i].device_id == therm[g_thermostats[i].device_id].device_id then
							for k,v in pairs(therm[g_thermostats[i].device_id]) do
								if v ~= g_thermostats[i][k] then
									g_thermostats[i][k] = v
									flagThermostat = true
									debug("getStreamEvents", "updating g_thermostats[" .. i .. "][" .. k .. "] with value =" .. tostring(v))
								end
							end
						end
					end
				end
				debug("getStreamEvents", "Verifying Smoke and CO detector/s!")
				local smokePoll = response.data.smoke_co_alarms
				--[[
				for key, val in pairs(smokePoll) do
					for k,v in pairs(val) do
						debug("getStreamEvents","smokePoll[" .. key .."][" .. tostring(k) .. "]=[" .. tostring(v) .. "]")
					end
				end	
				]]
				if smokePoll then
					for i=1,#g_smoke do
						if g_smoke[i].device_id == smokePoll[g_smoke[i].device_id].device_id then
							for k,v in pairs(smokePoll[g_smoke[i].device_id]) do
								if v ~= g_smoke[i][k] then
									g_smoke[i][k] = v
									flagSmoke = true
									debug("getStreamEvents", "updating g_smoke[" .. i .. "][" .. k .. "] with value =" .. tostring(v))
								end
							end
						end
					end
				end
			end -- if newEvent
			if flagThermostat then
				updateThermostatValues()
			end
			if flagSmoke then
				--updateSmokeValues()
			end
		end -- if fileSize > g_fileSize
	end
	-- check file size and delete if needed
	if tonumber(fileSize) > g_maxFileSize then
		-- delete /tmp/NestOutput.txt file & restart streaming
		startStreaming()
		debug("getStreamEvents", "Restarting Streaming!")
	end
	luup.call_delay("getStreamEvents", POLLING_RATE)
end

local function getNestData()
	local url = g_URL .. "?auth=" .. nest_token
	local header = "-H 'Accept: application/json'"
	local response, response_string = nestRequest("GET", header, url, "")
	if not response then
		debug("getNestData", "ERROR : Could not get Nest Data!")
	else
		debug("getNestData", "Nest Data = " .. response_string)
		if response.devices.thermostats then
			local count = 1
			g_thermostats[count] = {}
			for k,v in pairs(response.devices.thermostats) do
				g_thermostats[count].device_id = v.device_id
				g_thermostats[count].name = v.name
				g_thermostats[count].name_long = v.name_long
				g_thermostats[count].locale = v.locale
				g_thermostats[count].structure_id = v.structure_id
				g_thermostats[count].software_version = v.software_version
				g_thermostats[count].is_online = v.is_online
				g_thermostats[count].last_connection = v.last_connection
				g_thermostats[count].can_cool = v.can_cool
				g_thermostats[count].can_heat = v.can_heat
				g_thermostats[count].is_using_emergency_heat = v.is_using_emergency_heat
				g_thermostats[count].has_fan = v.has_fan
				g_thermostats[count].fan_timer_active = v.fan_timer_active
				g_thermostats[count].fan_timer_timeout = v.fan_timer_timeout
				g_thermostats[count].has_leaf = v.has_leaf
				g_thermostats[count].temperature_scale = v.temperature_scale
				g_thermostats[count].target_temperature_f = v.target_temperature_f
				g_thermostats[count].target_temperature_c = v.target_temperature_c
				g_thermostats[count].target_temperature_high_f = v.target_temperature_high_f
				g_thermostats[count].target_temperature_high_c = v.target_temperature_high_c
				g_thermostats[count].target_temperature_low_f = v.target_temperature_low_f
				g_thermostats[count].target_temperature_low_c = v.target_temperature_low_c
				g_thermostats[count].away_temperature_high_f = v.away_temperature_high_f
				g_thermostats[count].away_temperature_high_c = v.away_temperature_high_c
				g_thermostats[count].away_temperature_low_f = v.away_temperature_low_f
				g_thermostats[count].away_temperature_low_c = v.away_temperature_low_c
				g_thermostats[count].hvac_mode = v.hvac_mode
				g_thermostats[count].ambient_temperature_f = v.ambient_temperature_f
				g_thermostats[count].ambient_temperature_c = v.ambient_temperature_c
				g_thermostats[count].humidity = v.humidity
				count = count + 1
			end
		end
		if response.devices.smoke_co_alarms then
			local count = 1
			g_smoke[count] = {}
			for k,v in pairs(response.devices.smoke_co_alarms) do
				g_smoke[count].name = v.name
				g_smoke[count].name_long = v.name_long
				g_smoke[count].locale = v.locale
				g_smoke[count].structure_id = v.structure_id
				g_smoke[count].software_version = v.software_version
				g_smoke[count].device_id = v.device_id
				g_smoke[count].is_online = v.is_online
				g_smoke[count].last_connection = v.last_connection
				g_smoke[count].battery_health = v.battery_health
				g_smoke[count].co_alarm_state = v.co_alarm_state
				g_smoke[count].smoke_alarm_state = v.smoke_alarm_state
				g_smoke[count].ui_color_state = v.ui_color_state
				count = count + 1
			end
		end
		if response.devices.structures then
			local count = 1
			g_structures[count] = {}
			for k,v in pairs(response.structures) do
				g_structures[count].structure_id = v.structure_id
				g_structures[count].thermostats = v.thermostats
				g_structures[count].smoke_co_alarms = v.smoke_co_alarms
				g_structures[count].name = v.name
				g_structures[count].away = v.away
				-- [[
				g_structures[count].is_online = v.is_online
				g_structures[count].last_connection = v.last_connection
				g_structures[count].battery_health = v.battery_health
				g_structures[count].co_alarm_state = v.co_alarm_state
				g_structures[count].smoke_alarm_state = v.smoke_alarm_state
				g_structures[count].ui_color_state = v.ui_color_state
				-- ]]
				count = count + 1
			end
		end
	end
end

local function loginToNest(device, username, password)
	
	local url = "https://home.nest.com/user/login"
	local curlCommand = "curl -k --data 'username=" .. g_username .. "&password=" ..g_password .. "' https://home.nest.com/user/login"
	log("loginToNest", "curlCommand = " .. tostring(curlCommand))
	local stdout = io.popen(curlCommand)
	local response = stdout:read("*a")
	stdout:close()	
	log("loginToNest", "curlCommand Response = " .. tostring(response))
	local json_response = dkjson.decode(response)
	local data = {}
	if json_response.error then
		log("loginToNest", "error = " .. json_response.error)
		log("loginToNest", "description = " .. json_response.error_description)
    else
		data.transport_url = json_response.urls.transport_url
        data.userid = json_response.userid
        data.access_token = json_response.access_token
    end
	log("loginToNest","data.transport_url = [" .. data.transport_url .. "]" )
	log("loginToNest","data.userid = [" .. data.userid .. "]" )
	log("loginToNest","data.access_token = [" .. data.access_token .. "]" )
	url = "https://developer-api.nest.com/?auth=" .. data.access_token
	log("loginToNest","URL = [" .. url .. "]" )
	curlCommand = "curl -k -L -X GET -H 'Accept: application/json '" .. url
	log("loginToNest", "curlCommand2 = " .. tostring(curlCommand))
	stdout = io.popen(curlCommand)
	response = stdout:read("*a")
	stdout:close()	
	log("loginToNest", "curlCommand2 Response = " .. tostring(response))
	local json_response = dkjson.decode(response)
end
---------------------------------------------------------
---------------System Command Group----------------------
---------------------------------------------------------
local function updateThermostatValues()
	debug("updateThermostatValues","Verifying... ")
	local CurrentTemperature = 0
	local CurrentSetpoint = 0
	local SetpointTarget = 0
	local ModeStatus = ""
	local CurrentHumidity = 0
	for key, value in pairs(g_thermostats) do
		CurrentTemperature = tostring(value["ambient_temperature_" .. temperatureScale])
		CurrentSetpoint = tostring(value["target_temperature_" .. temperatureScale])
		ModeStatus = tostring(value.hvac_mode)
		CurrentHumidity = value.humidity
		if ModeStatus == "heat" then
			ModeStatus = "HeatOn"
		elseif ModeStatus == "cool" then
			ModeStatus = "CoolOn"
		elseif ModeStatus == "heat-cool" then
			ModeStatus = "AutoChangeOver"
		elseif ModeStatus == "off" then
			ModeStatus = "Off"
		end
		
		local AllSetpoints = CurrentSetpoint .. "," .. CurrentSetpoint .. ",0"
		--debug("updateThermostatValues","CurrentTemperature = " .. tostring(CurrentTemperature) .. " CurrentSetpoint = " .. tostring(CurrentSetpoint) .. " AllSetpoints = " .. tostring(AllSetpoints))
		--debug("updateThermostatValues","DEVICE ID = " .. tostring(value.vera_id))
		
		local ModeStatusUI = luup.variable_get("urn:upnp-org:serviceId:HVAC_UserOperatingMode1", "ModeStatus", tonumber(value.vera_id)) or ""
		local CurrentSetpointUI = luup.variable_get("urn:upnp-org:serviceId:TemperatureSetpoint1", "CurrentSetpoint", tonumber(value.vera_id)) or ""
		local CurrentTemperatureUI = luup.variable_get("urn:upnp-org:serviceId:TemperatureSensor1", "CurrentTemperature", tonumber(value.vera_id)) or ""
		local HumidityUI = luup.variable_get("urn:micasaverde-com:serviceId:HumiditySensor1", "CurrentLevel", tonumber(value.vera_id)) or ""
		
		if ModeStatus ~= ModeStatusUI then
			luup.variable_set("urn:upnp-org:serviceId:HVAC_UserOperatingMode1", "ModeStatus", ModeStatus, tonumber(value.vera_id))
		end
		if CurrentSetpoint ~= CurrentSetpointUI then
			luup.variable_set("urn:upnp-org:serviceId:TemperatureSetpoint1", "CurrentSetpoint", CurrentSetpoint, tonumber(value.vera_id))
			luup.variable_set("urn:upnp-org:serviceId:TemperatureSetpoint1", "AllSetpoints", AllSetpoints, tonumber(value.vera_id))
		end
		if CurrentTemperature ~= CurrentTemperatureUI then
			luup.variable_set("urn:upnp-org:serviceId:TemperatureSensor1", "CurrentTemperature", CurrentTemperature, tonumber(value.vera_id))
		end
		if tonumber(HumidityUI) ~= CurrentHumidity then
			luup.variable_set("urn:micasaverde-com:serviceId:HumiditySensor1", "CurrentLevel", CurrentHumidity ,tonumber(value.humidityDevice_id))
		end
	end
end

local function updateSmokeValues()

end
---------------------------------------------------------
---------------Action Implementations--------------------
---------------------------------------------------------
function clearTask()
	if (os.time() - g_lastTask >= DISPLAY_SECONDS) then
		luup.task("Clearing...", TASK.SUCCESS, "BoschPlugin", g_taskHandle)
	end
	debug("clearTask", "Clearing task... ")
end

function setModeTarget(device, mode)
	local VeraMode = mode
	if mode == "HeatOn" then 
		mode = "heat"
	elseif mode == "CoolOn" then
		mode = "cool"
	elseif mode == "AutoChangeOver" then
		mode = "heat-cool"
	elseif mode == "Off" then
		mode = "off"
	end
	
	local thermostatID = getNestDeviceID(device, g_thermostats)
	
	--curl -v -L -X PUT "https://developer-api.nest.com/structures/g-9y-2xkHpBh1MGkVaqXOGJiKOB9MkoW1hhYyQk2vAunCK8a731jbg?auth=<AUTH_TOKEN>" -H "Content-Type: application/json" -d '{"away":"away"}'
	local url = g_URL .. "/devices/thermostats/" .. tostring(thermostatID) .. "?auth=" .. nest_token
	local header = "-H 'Content-Type: application/json'"
	local data = "-d '{\"hvac_mode\" : \"" .. mode .."\" }'"
	log("setModeTarget", "data = " .. data)
	local response, response_string = nestRequest("PUT", header, url, data)
	if not response then
		log("setModeTarget", "ERROR : Could not set Data!")
	else
		log("setModeTarget", "Nest Response = " .. response_string)
		luup.variable_set("urn:upnp-org:serviceId:HVAC_UserOperatingMode1", "ModeStatus", VeraMode, device)
	end
end

function setCurrentSetpoint(device, value)
	local thermostatID = getNestDeviceID(device, g_thermostats)
	
	if tonumber(value) > 32 then
		value = "32"
	end
	--curl -v -L -X PUT "https://developer-api.nest.com/structures/g-9y-2xkHpBh1MGkVaqXOGJiKOB9MkoW1hhYyQk2vAunCK8a731jbg?auth=<AUTH_TOKEN>" -H "Content-Type: application/json" -d '{"away":"away"}'
	local url = g_URL .. "/devices/thermostats/" .. tostring(thermostatID) .. "?auth=" .. nest_token
	local header = "-H 'Content-Type: application/json'"
	local data = "-d '{\"target_temperature_" .. temperatureScale .. "\" :" .. tonumber(value) .. "}'"
	log("setCurrentSetpoint", "data = " .. data)
	local response, response_string = nestRequest("PUT", header, url, data)
	if not response then
		log("setCurrentSetpoint", "ERROR : Could not set Data!")
	else
		log("setCurrentSetpoint", "Request OK, Response = " .. response_string)
		local all = luup.variable_get("urn:upnp-org:serviceId:TemperatureSetpoint1", "AllSetpoints", device) or ""
		local now = luup.variable_get("urn:upnp-org:serviceId:TemperatureSetpoint1", "CurrentSetpoint", device) or ""
		all = string.gsub( all, now, value, 1 )
		log("setCurrentSetpoint", "ALL = [" .. all .. "]")
		luup.variable_set("urn:upnp-org:serviceId:TemperatureSetpoint1", "AllSetpoints", all, device)
		luup.variable_set("urn:upnp-org:serviceId:TemperatureSetpoint1", "CurrentSetpoint", value, device)
	end
end

function setArmed(device, value)
	luup.variable_set(SID.SES, "Armed", value, device)
end

function setTarget(device, value)
	luup.variable_set(SID.SWP, "Status", value, device)
	luup.variable_set(SID.SWP, "Target", value, device)
end

function connectWithNest(device, username, password)
	log("connectWithNest", "AM AJUNS!!!!")
	loginToNest(device, username, password)
end
---------------------------------------------------------
---------------Initialization Functions------------------
---------------------------------------------------------
local function getChildDevices(device)
	for dev, attr in pairs(luup.devices) do
		if (attr.device_num_parent == device) then
			local thermostatID = attr.id:match("^nest_thermostat_(.+)")
			local smokeID = attr.id:match("^nest_sco_(.+)")
			local humidity = attr.id:match("^nest_humidity_(.+)")
			if thermostatID then
				for k,v in pairs(g_thermostats) do
					if thermostatID == tostring(v.device_id) then
						g_thermostats[tonumber(k)].vera_id = dev
					end
				end
			end
			if smokeID then
				for k,v in pairs(g_smoke) do
					if smokeID == tostring(v.device_id) then
						g_smoke[tonumber(k)].vera_id = dev
					end
				end
			end
			if humidity then
				for k,v in pairs(g_thermostats) do
					if humidity == tostring(v.device_id) then
						g_thermostats[tonumber(k)].humidityDevice_id = dev
					end
				end
			end
		end
	end
end


local function appendDevices()
	local count  = 0
	for k, v in ipairs(g_thermostats) do
		debug("appendDevices", "Appending thermostat ".. v.name ..".")
		luup.chdev.append(lug_device, g_appendPtr, "nest_thermostat_".. v.device_id, v.name, DEVICE_TYPES.THERMOSTAT, DEVICE_FILES.THERMOSTAT, nil, nil, false)
		debug("appendDevices", "Appending Home/Away ".. v.name ..".")
		luup.chdev.append(lug_device, g_appendPtr, "nest_home_away_".. v.device_id, v.name, DEVICE_TYPES.BINARY_LIGHT, DEVICE_FILES.BINARY_LIGHT, nil, nil, false)
		if v.humidity then
			debug("appendDevices", "Appending humidity ".. v.name ..".")
			luup.chdev.append(lug_device, g_appendPtr, "nest_humidity_".. v.device_id, "Humidity - " ..v.name, DEVICE_TYPES.HUMIDITY, DEVICE_FILES.HUMIDITY, nil, "urn:micasaverde-com:serviceId:HumiditySensor1,CurrentLevel=" .. v.humidity, false)
		end
		count = count + 1
		if count > 20 then
			debug("appendDevices", "Possible error in generating outputs, more then 20 devices were generated!!!")
			return
		end
	end
	
	for k, v in ipairs(g_smoke) do
		debug("appendDevices", "Appending smoke sensor ".. v.name ..".")
		luup.chdev.append(lug_device, g_appendPtr, "nest_sco_".. v.device_id, v.name, DEVICE_TYPES.SMOKE_SENSOR, DEVICE_FILES.SMOKE_SENSOR, nil, nil, false)
		count = count + 1
		if count > 20 then
			debug("appendDevices", "Possible error in generating outputs, more then 20 devices were generated!!!")
			return
		end
	end
end

local function getInfos(device)
	
	local flagError = false
	
	local debugMode = luup.variable_get(SID.NEST, "DebugMode", device) or ""
	if (debugMode ~= "") then
		DEBUG_MODE = (debugMode == "1") and true or false
	else
		luup.variable_set(SID.NEST, "DebugMode", (DEBUG_MODE and "1" or "0"), device)
	end
	log("getInfos", "Debug mode "..(DEBUG_MODE and "enabled" or "disabled")..".")

	local polling_rate = luup.variable_get(SID.NEST, "POLLING_RATE", device) or ""
	if (polling_rate ~= "") then
		POLLING_RATE = tonumber(polling_rate)
	else
		luup.variable_set(SID.NEST, "POLLING_RATE", POLLING_RATE, device)
	end
	debug("getInfos", "POLLING_RATE = " .. POLLING_RATE )
	
	local username = luup.devices[device].user or ""
	local password = luup.devices[device].pass or ""
	
	getTemperatureScale()
	log("getInfos", "Temperature Scale is = " .. temperatureScale)
	
	if username == "" then
		luup.attr_set("username","default",device)
		g_username = "default"
		flagError = true
	else
		if username == "default" then
			flagError = true
		else 
			g_username = username
		end
	end

	if password == "" then
		luup.attr_set("password","default",device)
		g_password = "default"
		flagError = true
	else
		if password == "default" then
			flagError = true
		else
			g_password = password
		end
	end
	
	local token = luup.variable_get(SID.NEST, "TOKEN", device) or ""
	if token ~= "" and token ~= "default" then
		nest_token = token
	else
		luup.variable_set(SID.NEST, "TOKEN", "default", device)
	end
	return flagError
end

function Init(lul_device)
	lug_device = lul_device
	g_taskHandle = luup.task("Starting up...", TASK.BUSY, "Vera Connect WWN", g_taskHandle)
	local continue = getInfos(lug_device)
	if continue then
		return false, "Username and/or password not set. Please add them in 'Advance' tab!", "Vera Connect WWN"
	end
	getNestData()
	g_appendPtr = luup.chdev.start(lug_device)
	appendDevices(lug_device)
	luup.chdev.sync(lug_device, g_appendPtr)
	getChildDevices(lug_device)
	log("Init", "Updating thermostat values!")
	updateThermostatValues()
	log("Init", "Trying to start streaming and get the latest event/s!")
	startStreaming()
	getStreamEvents()
	-------------------------------
	for i=1,#g_thermostats do
		for k,v in pairs(g_thermostats[i]) do
			log("Init","g_thermostats[".. i .."].[" .. tostring(k) .. "]=[" .. tostring(v) .. "]")
		end
	end	
	for i=1,#g_smoke do
		for k,v in pairs(g_smoke[i]) do
			log("Init","g_smoke[".. i .."].[" .. tostring(k) .. "]=[" .. tostring(v) .. "]")
		end
	end	
	for i=1,#g_structures do
		for k,v in pairs(g_structures[i]) do
			log("Init","g_structures[".. i .."].[" .. tostring(k) .. "]=[" .. tostring(v) .. "]")
		end
	end	
	------------------------------
	log( "Startup", "Startup Successful" )
	g_taskHandle = luup.task("Startup successful.", TASK.BUSY, "Vera Connect WWN", g_taskHandle)
	luup.set_failure(0, lul_device)
	return true, "Startup successful.", "Vera Connect WWN"
end
